/*
 * File:    crypt.c
 * Author:  Holger Klawitter
 */

static crypt_globals_t *
InitCryptGlobals(void)
{
  crypt_globals_t *crg;
  
  if ((crg = MemPtrNew(sizeof(crypt_globals_t))))
    crg->ks = NULL;
  return crg;
}


static void
FreeCryptGlobals(crypt_globals_t *crg)
{
  if (crg) {
    if (crg->ks) {
      MemSet(crg->ks, sizeof(UInt16)*INTERNAL_KEY_LENGTH, 0);
      MemPtrFree(crg->ks);
    }
    MemPtrFree(crg);
  }
}


//////////////////////////////////////////////////////////////////////
// The IDEA(tm) algorithm is copyrighted and trademarked by Ascom
// The folloing implementation is copyrighted by the
// Cryptics Development Team (Systemics, Ltd.)
//////////////////////////////////////////////////////////////////////

static UInt16
mul(UInt32 a, UInt32 b)
{
  UInt32 p;
  UInt16 res;

  a &= 0xFFFF;
  b &= 0xFFFF;
  if (a) {
    if (b) {
      p = a * b;
      b = p & 0xFFFF;
      a = p >> 16;
      res = (UInt16)(b - a + (b < a ? 1 : 0));
    } else
      res = (UInt16)(1 - a);
  } else
    res = (UInt16)(1 - b);

  return res;
}


static UInt16
inv(Int32 xx)
{
  Int32 x,t1,y,t0,q;

  x = xx & 0xFFFF;     // only lower 16 bits
  if (x <= 1)
    return (UInt16)x;  // 0 and 1 are self-inverse
  
  t1 = (UInt32)0x10001 / x;  // Since x >= 2, this fits into 16 bits
  y = (UInt32)0x10001 % x;
  if (y == 1)
    return (UInt16)(1 - t1);
  
  t0 = 1;
  do {
    q = x / y;
    x = x % y;
    t0 += q * t1;
    if (x == 1)
      return (UInt16)t0;
    q = y / x;
    y %= x;
    t1 += q * t0;
  } while (y != 1);

  return (UInt16)(1 - t1);
}


static void
loadKey(UInt8 *userkey, crypt_globals_t *crg)
{
  Int16 i, j;
  Int16 zoff;

  if (!crg->ks)
    crg->ks = (UInt16*)MemPtrNew(sizeof(UInt16)*INTERNAL_KEY_LENGTH);

  for (i=0, j=0; i<8; i++, j += 2)
    crg->ks[i] = ((UInt16)(userkey[j] & 0xFF) << 8 | (userkey[j+1] & 0xFF));
  
  for (i = 0, zoff = 0, j = 8; j < INTERNAL_KEY_LENGTH; i &= 7, j++) {
    i++;
    crg->ks[i + 7 + zoff] =
      (UInt16)(((UInt32)(crg->ks[(i & 7) + zoff]) << 9) |
      (((UInt32)(crg->ks[((i + 1) & 7) + zoff]) >> 7) & 0x1FF));
    zoff += i & 8;
  }
}


static void
invertKey(crypt_globals_t *crg)
{
  int i, j = 4, k = INTERNAL_KEY_LENGTH - 1;
  UInt16 temp[INTERNAL_KEY_LENGTH];

  temp[k--] = inv(crg->ks[3]);
  temp[k--] = (UInt16) (-1)*(crg->ks[2]);
  temp[k--] = (UInt16) (-1)*(crg->ks[1]);
  temp[k--] = inv(crg->ks[0]);
  for (i = 1; i < ROUNDS; i++, j += 6) {
    temp[k--] = crg->ks[j + 1];
    temp[k--] = crg->ks[j];
    temp[k--] = inv(crg->ks[j + 5]);
    temp[k--] = (UInt16) (-1)*(crg->ks[j + 3]);
    temp[k--] = (UInt16) (-1)*(crg->ks[j + 4]);
    temp[k--] = inv(crg->ks[j + 2]);
  }
  temp[k--] = crg->ks[j + 1];
  temp[k--] = crg->ks[j];
  temp[k--] = inv(crg->ks[j + 5]);
  temp[k--] = (UInt16) (-1)*(crg->ks[j + 4]);
  temp[k--] = (UInt16) (-1)*(crg->ks[j + 3]);
  temp[k--] = inv(crg->ks[j + 2]);
  MemMove(crg->ks, temp, sizeof(UInt16)*INTERNAL_KEY_LENGTH);
}


static void
blockEncrypt(UInt8 *text, crypt_globals_t *crg)
{
  UInt16
    x1 = (UInt16)(((text[0] & 0xFF) << 8) | (text[1] & 0xFF)),
    x2 = (UInt16)(((text[2] & 0xFF) << 8) | (text[3] & 0xFF)),
    x3 = (UInt16)(((text[4] & 0xFF) << 8) | (text[5] & 0xFF)),
    x4 = (UInt16)(((text[6] & 0xFF) << 8) | (text[7] & 0xFF)),
    s2, s3;
  int i = 0,
    round = ROUNDS;
  
  while (round-- > 0) {
    x1 = mul(x1, crg->ks[i++]);
    x2 += crg->ks[i++];
    x3 += crg->ks[i++];
    x4 = mul(x4, crg->ks[i++]);
    
    s3 = x3;
    x3 = mul(x1 ^ x3, crg->ks[i++]);
    s2 = x2;
    x2 = mul(x3 + (x2 ^ x4), crg->ks[i++]);
    x3 += x2;
    
    x1 ^= x2;
    x4 ^= x3;
    x2 ^= s3;
    x3 ^= s2;
  }
  s2 = mul(x1, crg->ks[i++]);
  text[0] = (UInt8)(s2 >> 8);
  text[1] = (UInt8) s2;
  s2 = (UInt16)(x3 + crg->ks[i++]);
  text[2] = (UInt8)(s2 >> 8);
  text[3] = (UInt8) s2;
  s2 = (UInt16)(x2 + crg->ks[i++]);
  text[4] = (UInt8)(s2 >> 8);
  text[5] = (UInt8) s2;
  s2 = mul(x4, crg->ks[i]);
  text[6] = (UInt8)(s2 >> 8);
  text[7] = (UInt8) s2;
}

//////////////////////////////////////////////////////////////////////
// Foreign code ends here.
//////////////////////////////////////////////////////////////////////


static void
preparePrimaryKey(Char *key, crypt_globals_t *crg)
{
  Char longerKey[KEY_LENGTH];
  
  if (key && StrLen(key) >= KEY_LENGTH) {
    loadKey(key, crg);
  } else {
    MemSet(longerKey, KEY_LENGTH, 0);
    if (key) StrCopy(longerKey, key);
    loadKey(longerKey, crg);
    MemSet(longerKey, KEY_LENGTH, 0); // dont leave clean text behind!
  }
}


static void
prepareKey(Char *key, Boolean weakEncrypt, crypt_globals_t *crg)
{
  Char *realKey;
  UInt16 keyLen;
  UInt16 realKeyLen;
  UInt16 padding;
  Int16 i,j;

  // create a weaker key to initiate encrypted "real" key.
  preparePrimaryKey(key, crg);
  if (weakEncrypt)
    return;

  // create clean text key with matching blocklen
  // and do some nice padding
  keyLen = StrLen(key);
  realKeyLen = (keyLen + KEY_LENGTH-1) & 0xfff0;

  padding = realKeyLen - keyLen;
  realKey = MemPtrNew(sizeof(Char)*realKeyLen);
  MemSet(realKey, realKeyLen, padding + (padding << 4));
  StrCopy(realKey, key);
  
  // encrypt key with weak key and cipher block chaining
  // BLOCK_SIZE must divide KEY_LENGTH
  for (i=0; i<=realKeyLen-BLOCK_SIZE; i+=BLOCK_SIZE) {
    if (i>0) {
      for (j=0; j<BLOCK_SIZE; j++) {
        realKey[i+j] ^= realKey[i+j-BLOCK_SIZE];
      }
    }
    blockEncrypt(realKey + i, crg);
  }

  // get the last part as the real key
  i -= KEY_LENGTH;
  loadKey(realKey+i, crg);

  // cleanup
  MemSet(realKey, realKeyLen, 0);
  MemPtrFree(realKey);
}


static void
MyStrNCopy(Char *dest, Char* src, UInt16 n)
{
  while (n-->0 && *src) {
    *dest++ = *src++;
  }
  if (n)
    *dest = 0;
}


static Char*
encrypt(Char *key,
        Char *text,
        Boolean weakEncrypt,
        crypt_globals_t *crg,
        armour_globals_t *ag)
{
  UInt8 lastBlock[BLOCK_SIZE];
  UInt8 *sText;
  Char  *aText;
  Int16 i,j;
  Int16 len, sLen;

  if (!text || !key ||
      !(len=StrLen(text)) ||
      !StrLen(key))
    return NULL;
  sLen = ((len+10)/8)*8; // + 3 bytes, round up to multiple of 8

  // bring text to blocklength and encode length of string into it
  sText = (UInt8*)MemPtrNew(sizeof(UInt8)*sLen);
  MemSet(sText, sLen, 0);
  MyStrNCopy(sText+3, text, len);
  sText[0] = len & 0xFF;
  sText[1] = len >> 8;
  sText[2] = SysRandom(0);

  prepareKey(key, weakEncrypt, crg);

  // for(i=0; i<sLen; i++) { printf("(%d)", sText[i]); } printf("\n");
  MemSet(lastBlock, BLOCK_SIZE, 0);
  for (i=0; i<sLen; i+=BLOCK_SIZE) {
    for (j=0; j<BLOCK_SIZE; j++)
      lastBlock[j] ^= sText[i+j];
    blockEncrypt(lastBlock, crg);
    MemMove(sText+i, lastBlock, BLOCK_SIZE);
  }
  //for(i=0; i<sLen; i++) { printf("(%d)", sText[i]); } printf("\n");

  aText = armour(sText, sLen, ag);
  MemSet(sText, sLen, 0);
  MemPtrFree(sText);
  return aText;
}


static Char*
decrypt(Char *key,
        Char *aText,
        Boolean weakEncrypt,
        crypt_globals_t *crg,
        armour_globals_t *ag)
{
  Char  lastBlock[BLOCK_SIZE];
  Char  curBlock[BLOCK_SIZE];
  UInt8 *sText;
  Char  *text;
  Int16 i,j;
  Int16 len,sLen;

  if (!aText || !key ||
      (StrLen(aText) < 2) ||
      !StrLen(key))
    return NULL;

  // remove armouring, now the text may contain zeros!
  sText = deArmour(aText, &sLen, ag);

  // load key into idea
  prepareKey(key, weakEncrypt, crg);
  invertKey(crg);

  //for(i=0; i<sLen; i++) { printf("(%d)", sText[i]); } printf("\n");
  MemSet(lastBlock, BLOCK_SIZE, 0);
  for (i=0; i<sLen; i+=BLOCK_SIZE) {
    MemMove(curBlock, sText+i, BLOCK_SIZE);
    blockEncrypt(sText+i, crg);
    for (j=0; j<BLOCK_SIZE; j++)
      sText[i+j] ^= lastBlock[j];
    MemMove(lastBlock, curBlock, BLOCK_SIZE);
  }

  // readout text in correct length
  len = (UInt8)(sText[1]) << 8 | (UInt8)sText[0];
  if (len>0 && len+3 <= sLen) {
    text = (Char*)MemPtrNew(sizeof(Char)*(len+1));
    MemMove(text, sText+3, len);
    text[len] = 0;
  } else
    text = NULL;

  // cleanup
  MemSet(sText, sLen, 0);
  MemPtrFree(sText);

  return text;
}

// EOF
